//! Socket related types and functions

use ::std::{net::SocketAddr, sync::Arc, vec::Vec};
use ::tokio::{net::UdpSocket, task::JoinHandle};

/// Pair to easily track the socket and its async listening handle
pub type SocketTracker =
    (Arc<UdpSocket>, Arc<JoinHandle<::std::io::Result<()>>>);

/// async free socket list
pub(crate) struct SocketList {
    pub list: Vec<SocketTracker>,
}
impl SocketList {
    pub(crate) fn new() -> Self {
        Self { list: Vec::new() }
    }
    pub(crate) fn rm_all(&mut self) -> Self {
        let mut old_list = Vec::new();
        ::core::mem::swap(&mut self.list, &mut old_list);
        Self { list: old_list }
    }
    pub(crate) async fn add_socket(
        &mut self,
        socket: Arc<UdpSocket>,
        handle: JoinHandle<::std::io::Result<()>>,
    ) {
        let arc_handle = Arc::new(handle);
        self.list.push((socket, arc_handle));
    }
    /// This method assumes no other `add_sockets` are being run
    pub(crate) async fn stop_all(self) {
        for (_socket, mut handle) in self.list.into_iter() {
            let _ = Arc::get_mut(&mut handle).unwrap().await;
        }
    }
}

/// Strong typedef for a client socket address
#[derive(Debug, Copy, Clone)]
pub(crate) struct UdpClient(pub SocketAddr);
/// Strong typedef for a server socket address
#[derive(Debug, Copy, Clone)]
pub(crate) struct UdpServer(pub SocketAddr);

/// Enable some common socket options. This is just the unsafe part
fn enable_sock_opt(
    fd: ::std::os::fd::RawFd,
    option: ::libc::c_int,
    value: ::libc::c_int,
) -> ::std::io::Result<()> {
    #[allow(unsafe_code)]
    unsafe {
        #[allow(trivial_casts)]
        let val = &value as *const _ as *const ::libc::c_void;
        let size = ::std::mem::size_of_val(&value) as ::libc::socklen_t;
        // always clear the error bit before doing a new syscall
        let _ = ::std::io::Error::last_os_error();
        let ret = ::libc::setsockopt(fd, ::libc::SOL_SOCKET, option, val, size);
        if ret != 0 {
            return Err(::std::io::Error::last_os_error());
        }
    }
    Ok(())
}
/// Add an async udp listener
pub async fn bind_udp(sock: SocketAddr) -> ::std::io::Result<UdpSocket> {
    let socket = UdpSocket::bind(sock).await?;

    use ::std::os::fd::AsRawFd;
    let fd = socket.as_raw_fd();
    // can be useful later on for reloads
    enable_sock_opt(fd, ::libc::SO_REUSEADDR, 1)?;
    enable_sock_opt(fd, ::libc::SO_REUSEPORT, 1)?;

    // We will do path MTU discovery by ourselves,
    // always set the "don't fragment" bit
    if sock.is_ipv6() {
        enable_sock_opt(fd, ::libc::IPV6_DONTFRAG, 1)?;
    } else {
        // FIXME: linux only
        enable_sock_opt(fd, ::libc::IP_MTU_DISCOVER, ::libc::IP_PMTUDISC_DO)?;
    }

    Ok(socket)
}
